from demogpt_agenthub.tools.base import BaseTool

class TavilySearchTool(BaseTool):
    def __init__(self, max_results=5, topic="general", search_depth="basic"):
        from langchain_tavily import TavilySearch
        self.tool = TavilySearch(
            max_results=max_results,
            topic=topic,
            search_depth=search_depth,
        )
        super().__init__()
        self.name = "TavilySearch"
        self.description = """A search engine tool that can find information on the internet.
        It is useful for answering questions about current events, facts, and general knowledge.
        It gets a search query as input and returns relevant search results."""

    def run(self, query: str):
        results = self.tool.invoke({"query": query})
        if isinstance(results, list):
            return "\n\n".join(
                r.get("content", str(r)) if isinstance(r, dict) else str(r)
                for r in results
            )
        return str(results)

if __name__ == "__main__":
    tool = TavilySearchTool()
    print(tool.run("What is the capital of France?"))



from collections import defaultdict
from demogpt_agenthub.tools.base import BaseTool

class YoloTool(BaseTool):
    def __init__(self):
        try:
            from ultralytics import YOLO
        except ImportError:
            raise ImportError(
                "YoloTool requires the 'ultralytics' package. "
                "Install it with: pip install ultralytics"
            )
        self.model = YOLO("yolo11n.pt")
        super().__init__()
        self.name = "YOLO Object Detection"
        self.description = """A tool that can execute YOLO object detection by using the YOLO model.
        It returns a detailed description of the detected objects in the image.
        It gets the path of the image as an argument."""
        
    def run(self, path):
        results = self.model(path)[0]
        boxes = results.boxes
        names = results.names
        classes = results.boxes.cls
        resulting_str = ""
        object2count = defaultdict(int)
        for box, cls in zip(boxes, classes):
            
            x1, y1, x2, y2 = map(int, box.xyxy[0].tolist())
            _class = names[int(cls)]
            object2count[_class] += 1
            resulting_str += f"The {_class}[{object2count[_class]}] is from ({x1}, {y1}) to ({x2}, {y2}).\n"
        intro = ""
        for obj, count in object2count.items():
            intro += f"{count} {obj}, "
        intro = intro.strip(", ")
        if len(intro) > 0:
            intro = "There are " + intro + " in the image.\n"
        else:
            intro = "There are no objects in the image.\n"
        resulting_str = intro + resulting_str
        if len(resulting_str) > 0:
            resulting_str = "in the image.\n" + resulting_str
        return resulting_str
    
if __name__ == "__main__":
    tool = YoloTool()
    path = "/home/melih/Downloads/bus.jpg"
    print(tool.run(path))


from langchain_community.tools import ShellTool
from demogpt_agenthub.tools import BaseTool
class BashTool(BaseTool):
    def __init__(self):
        self.tool = ShellTool()
        super().__init__()
    
    def run(self, commands):
        if isinstance(commands, str):
            return self.tool.run(({"commands": [commands]}))
        return self.tool.run(({"commands": commands}))
    
if __name__ == "__main__":
    commands = ["ls", "pwd"]
    tool = BashTool()
    print(tool.run(commands))
    command = "ls"
    print(tool.run(command))


from langchain_community.tools.wikidata.tool import WikidataAPIWrapper, WikidataQueryRun
from demogpt_agenthub.tools import BaseTool

class WikiDataTool(BaseTool):
    def __init__(self):
        self.tool = WikidataQueryRun(api_wrapper=WikidataAPIWrapper())
        super().__init__()

    def run(self, query):
        return self.tool.run(query)
    
if __name__ == "__main__":
    query = "Alan Turing"
    tool = WikiDataTool()
    print(tool.run(query))


system_template = """
You are a helpful assistant that is good at deciding if the context successfully includes the information needed to complete the task.
The context is a list of messages between the user and the agent.
The agent will sometimes call tools to get more information.
The called tools together with their results are also included in the context.

You will be given a task and the current context.
You need to decide if the context includes all the necessary information needed to complete the task.
You will first reason about the task and the context.
Then, you will give if there is any missing information.
If the task requires a computation, you must use the necessary tool provided to get the missing information.
For the computation heavy tasks, you must use the necessary tool provided to get the missing information.
If you think the there is no need to use any tool, respond with "<YES>".
If you think the you still need to call a tool, respond with "<NO>".
"""

human_template = """
Task: {task}

Tools: {tools}

Context: {context}

Your reasoning and decision (<YES> or <NO>):
"""


from langchain_community.tools import WikipediaQueryRun
from langchain_community.utilities import WikipediaAPIWrapper
from demogpt_agenthub.tools import BaseTool

class WikipediaTool(BaseTool):
    def __init__(self):
        self.tool = WikipediaQueryRun(api_wrapper=WikipediaAPIWrapper())
        super().__init__()

    def run(self, query):
        return self.tool.run(query)
    
if __name__ == "__main__":
    query = "How to create a list in python"
    tool = WikipediaTool()
    print(tool.run(query))


from langchain_community.utilities import ArxivAPIWrapper
from demogpt_agenthub.tools import BaseTool

class ArxivTool(BaseTool):
    def __init__(self):
        self.tool = ArxivAPIWrapper()
        super().__init__()

    def run(self, query):
        return self.tool.run(query)
    
if __name__ == "__main__":
    tool = ArxivTool()
    print(tool.run("2106.01495"))


from langchain_openai import ChatOpenAI
from langchain_core.prompts.chat import (ChatPromptTemplate,
                                         HumanMessagePromptTemplate,
                                         SystemMessagePromptTemplate)

def get_prompt(system_template: str, human_template: str):
    prompts = []
    if system_template:
        prompts.append(SystemMessagePromptTemplate.from_template(system_template))
    if human_template:
        prompts.append(HumanMessagePromptTemplate.from_template(human_template))
    return ChatPromptTemplate.from_messages(prompts)
    
if __name__ == "__main__":
    system_template = "System: {system_message}"
    human_template = "Human: {human_message}"
    prompt = get_prompt(system_template, human_template)
    print(prompt)


from langchain_community.utilities import TextRequestsWrapper
from demogpt_agenthub.tools import BaseTool

class RequestUrlTool(BaseTool):
    def __init__(self):
        self.tool = TextRequestsWrapper()
        super().__init__()
        
    def run(self, url):
        return self.tool.get(url)
    
if __name__ == "__main__":
    tool = RequestUrlTool()
    print(tool.run("https://www.google.com"))


from demogpt_agenthub.agents import BaseAgent

class ToolCallingAgent(BaseAgent):
    def __init__(self, tools, llm, verbose=False):
        super().__init__(tools, llm, verbose)
    
    def run(self, prompt):
        self.add_message("User", prompt)
        # Format tool explanations for better readability
        formatted_tools = "\nAvailable Tools:\n"
        for tool_name, details in self.tool_explanations.items():
            formatted_tools += f"\n{tool_name}:\n"
            formatted_tools += f"  Description: {details['description']}\n"
            formatted_tools += f"  {details['parameters']}\n"
            formatted_tools += f"  Run Function: {details['run_description']}\n"
        decision = self.tool_decider.invoke({"task": prompt, "context": self.context, "tools": formatted_tools})
        self.add_message("Agent", decision["reasoning"])
        self.pretty_print("Reasoning", decision["reasoning"])
        self.pretty_print("Tool call", decision["tool"])
        tool_call = self.tools[decision["tool"]]
        tool_args = decision["argument"]
        tool_result = tool_call.run(**tool_args)
        self.add_message(decision["tool"], tool_result)
        self.pretty_print("Tool result", tool_result)
        answer = self.final_answer.invoke({"query": prompt, "context": self.context})
        self.add_message("Agent", answer)
        self.pretty_print("Answer", answer)
        return answer

if __name__ == "__main__":
    from demogpt_agenthub.tools import TavilySearchTool, WeatherTool, WikipediaTool
    from demogpt_agenthub.llms import OpenAIChatModel
    search_tool = TavilySearchTool()
    weather_tool = WeatherTool()
    wikipedia_tool = WikipediaTool()
    agent = ToolCallingAgent(tools=[search_tool, weather_tool, wikipedia_tool], llm=OpenAIChatModel(model_name="gpt-4o-mini"), verbose=True)
    query = "Who is Daron Acemoglu?"
    print(agent.run(query))


from langchain_core.documents import Document
from langchain_core.output_parsers import StrOutputParser
from langchain_core.runnables import RunnablePassthrough
from langchain_core.prompts import ChatPromptTemplate
from langchain_community.document_loaders import (
    TextLoader, 
    WebBaseLoader, 
    PyPDFLoader,
    DirectoryLoader,
    JSONLoader
    )
from langchain_community.document_loaders.csv_loader import CSVLoader
from typing import List, Union
import os
import shutil
import logging

from demogpt_agenthub.llms.base import BaseLLM
from demogpt_agenthub.prompts.rag.base import template

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class BaseRAG:
    OPENAI_EMBEDDING_MODELS = [
        "text-embedding-3-small",
        "text-embedding-3-large",
        "text-embedding-ada-002"
    ]
    def __init__(self, llm: BaseLLM, 
                 vectorstore: str, 
                 persistent_path: str, 
                 index_name: str,
                 reset_vectorstore: bool = False,
                 embedding_model_name: str = "sentence-transformers/all-mpnet-base-v2",
                 filter: dict = None,
                 k: int = 4,
                 verbose: bool = False):
        self.name = "RAG"
        self.description = "A tool that can search for information in vector database by using a query."
        self.llm = llm
        self.persistent_path = persistent_path
        self.verbose = verbose
        self.index_name = index_name
        self.search_kwargs={
            'k': k,
            'filter': filter
        }
        self.load_embedding_model(embedding_model_name)
        self.load_vectorstore(vectorstore, reset_vectorstore)
        self.output_parser = StrOutputParser()
        
        prompt = ChatPromptTemplate.from_template(template)

        self.rag_chain = (
            {"context": self.retriever, "question": RunnablePassthrough()}
            | prompt
            | self.llm
            | self.output_parser
        )

    def load_embedding_model(self, model_name: str):
        if model_name in self.OPENAI_EMBEDDING_MODELS:
            from langchain_openai import OpenAIEmbeddings
            self.embedding_model = OpenAIEmbeddings(model=model_name)
        else:
            from langchain_huggingface import HuggingFaceEmbeddings
            self.embedding_model = HuggingFaceEmbeddings(model_name=model_name)

    def load_vectorstore(self, vectorstore, reset_vectorstore: bool = False):
        # First, close any existing connection
        if hasattr(self, 'vectorstore') and hasattr(self.vectorstore, '_client'):
            self.vectorstore._client.close()
        
        if reset_vectorstore:
            if os.path.exists(self.persistent_path):
                print(f"Removing existing vectorstore at {self.persistent_path}")
                shutil.rmtree(self.persistent_path)
        if vectorstore == "chroma":
            from langchain_chroma import Chroma
            self.vectorstore = Chroma(embedding_function=self.embedding_model, persist_directory=self.persistent_path)
        elif vectorstore == "pinecone":
            from langchain_pinecone import PineconeVectorStore
            self.vectorstore = PineconeVectorStore(index_name=self.index_name)
        elif vectorstore == "faiss":
            from langchain_community.vectorstores import FAISS
            self.vectorstore = FAISS(self.persistent_path, self.embedding_model)
        else:
            raise ValueError(f"Vectorstore {vectorstore} not supported")
        
        self.retriever = self.vectorstore.as_retriever()
        
    def _add_documents(self, documents: List[Document]):
        self.vectorstore.add_documents(documents)

    def add_texts(self, texts: Union[str, List[str]]):
        if isinstance(texts, str):
            texts = [texts]
        self.vectorstore.add_texts(texts)
    
    def add_files(self, file_paths  : Union[str, List[str]]):
        if isinstance(file_paths, str):
            file_paths = [file_paths]
        docs = []
        for file_path in file_paths:  
            if file_path.endswith(".txt"):
                loader = TextLoader(file_path)
            elif file_path.endswith(".pdf"):
                loader = PyPDFLoader(file_path)
            elif file_path.endswith(".json"):
                loader = JSONLoader(file_path)
            elif file_path.endswith(".csv"):
                loader = CSVLoader(file_path)
            elif file_path.startswith("http"):
                loader = WebBaseLoader(file_path)
            else:
                logger.info(f"File {file_path} not supported")
                continue
            try:    
                docs.extend(loader.load())
            except Exception as e:
                logger.error(f"Error loading file {file_path}: {e}")
        self._add_documents(docs)

    def run(self, query: str):
        return self.rag_chain.invoke(query)

if __name__ == "__main__":
    from demogpt_agenthub.llms.openai import OpenAIChatModel
    rag = BaseRAG(llm=OpenAIChatModel(model="gpt-4o-mini"), 
                  vectorstore="chroma", 
                  persistent_path="rag_chroma", 
                  index_name="rag_index",
                  reset_vectorstore=True,
                  embedding_model_name="sentence-transformers/all-mpnet-base-v2",
                  filter={"search_kwargs": {"score_threshold": 0.5}}
                  )
    rag.add_files(["/home/melih/Downloads/results - 2024-11-04T160209.186.pdf"])
    print(rag.query("How many people joined the interview with Age 40 to 49?"))



from langchain_core.exceptions import OutputParserException
from langchain_core.output_parsers import BaseOutputParser


# The [bool] desribes a parameterization of a generic.
# It's basically indicating what the return type of parse is
# in this case the return type is either True or False
class BooleanOutputParser(BaseOutputParser[bool]):
    """Custom boolean parser."""

    true_val: str = "<YES>"
    false_val: str = "<NO>"

    def parse(self, text: str) -> bool:
        cleaned_text = text.strip().upper()
        if self.true_val.upper() in cleaned_text:
            return True
        elif self.false_val.upper() in cleaned_text:
            return False
        else:
            raise OutputParserException(
                f"BooleanOutputParser expected output value to either be "
                f"{self.true_val} or {self.false_val} (case-insensitive). "
                f"Received {cleaned_text}."
            )

    @property
    def _type(self) -> str:
        return "boolean_output_parser"


import os
from langchain_community.utilities import OpenWeatherMapAPIWrapper
from langchain_community.tools.openweathermap.tool import OpenWeatherMapQueryRun
from demogpt_agenthub.tools import BaseTool

class WeatherTool(BaseTool):
    def __init__(self):
        api_key = os.environ.get('OPENWEATHERMAP_API_KEY')
        if not api_key:
            raise ValueError('OPENWEATHERMAP_API_KEY environment variable is not set')
        wrapper = OpenWeatherMapAPIWrapper()
        self.tool = OpenWeatherMapQueryRun(api_wrapper=wrapper)
        super().__init__()
    
    def run(self, city):
        return self.tool.run(city)
    
if __name__ == '__main__':
    tool = WeatherTool()
    print(tool.run('London'))